import numpy
import numpy.ma as ma
import pandas
from numpy.lib.histograms import _hist_bin_selectors
from numpy.ma.core import nomask as _nomask

from .colors import hexcolor


def robust_min_max(
    s, threshold_for_dropping_zeros=0.35, cut_after_stdevs=5, percentiles_to_cut=1
):
    # Get s as an array (it should be already though)
    s = numpy.asarray(s)
    # mask out infs and nans
    sx = ma.masked_array(s, mask=~numpy.isfinite(s))
    n_values = sx.count()
    # Check if data is mostly zeros
    n_zeros = (s == 0).sum()
    if n_zeros / n_values > threshold_for_dropping_zeros:
        sx.mask |= sx == 0
        zeros_are_dropped = True
        n_values = sx.count()
    else:
        zeros_are_dropped = False
    if n_values == 0:
        # no data is left, default values
        bottom, top, range_is_truncated = 0, 1, 100
    else:
        mean = sx.mean()
        stdev = sx.std()
        bottom = sx.min()
        top = sx.max()
        range_is_truncated = 0
        cut_bottom = bottom < mean - cut_after_stdevs * stdev
        cut_top = top > mean + cut_after_stdevs * stdev
        if percentiles_to_cut > 0:
            if cut_bottom:
                bottom_ = numpy.percentile(sx, percentiles_to_cut)
                if not numpy.isnan(bottom_):
                    bottom = bottom_
                    range_is_truncated += percentiles_to_cut
            if cut_top:
                top_ = numpy.percentile(sx, 100 - percentiles_to_cut)
                if not numpy.isnan(top_):
                    top = top_
                    range_is_truncated += percentiles_to_cut
    return bottom, top, zeros_are_dropped, range_is_truncated, sx


def fixed_min_max(s, fixed_min, fixed_max, threshold_for_dropping_zeros=0.35):
    # Get s as an array (it should be already though)
    s = numpy.asarray(s)
    # mask out infs and nans
    sx = ma.masked_array(s, mask=numpy.isnan(s))
    n_values_finite = n_values = sx.count()

    n_zeros = (s == 0).sum()
    if n_zeros / n_values > threshold_for_dropping_zeros:
        sx.mask |= sx == 0
        zeros_are_dropped = True
        n_values = sx.count()
    else:
        zeros_are_dropped = False

    # Over max
    if fixed_max is not None:
        sx.mask |= sx >= fixed_max
    n_values_overmax = n_values - sx.count()
    n_values = sx.count()

    # Under Min
    if fixed_min is not None:
        sx.mask |= sx <= fixed_min
    n_values_undermin = n_values - sx.count()
    n_values = sx.count()

    pct_over = n_values_overmax / n_values_finite
    pct_under = n_values_undermin / n_values_finite

    return sx, pct_under, pct_over, zeros_are_dropped


def get_histogram_bins(
    data_to_bin,
    bins="sturges",
    percentiles_to_cut=1,
    maxrange=None,
    fixedrange=None,
    threshold_for_dropping_zeros=0.35,
):
    if fixedrange is None:
        bottom, top, zeros_are_dropped, range_is_truncated, sx = robust_min_max(
            data_to_bin, percentiles_to_cut=percentiles_to_cut
        )
        pct_under, pct_over = 0, 0
    else:
        bottom, top = fixedrange[0], fixedrange[1]
        sx, pct_under, pct_over, zeros_are_dropped = fixed_min_max(
            data_to_bin,
            bottom,
            top,
            threshold_for_dropping_zeros=threshold_for_dropping_zeros,
        )
        range_is_truncated = (pct_under + pct_over) * 100
    bottom = float(bottom) if bottom is not None else sx.min() - 1e-8
    top = float(top) if top is not None else sx.max() + 1e-8

    if numpy.isinf(bottom):
        bottom = sx.min()
    if numpy.isinf(top):
        top = sx.max()

    if maxrange is not None:
        if maxrange[0] > bottom:
            bottom = float(maxrange[0])
        if maxrange[1] < top:
            top = float(maxrange[1])

    if bottom > top:  # unclear how this would happen, but just in case
        bottom, top = top, bottom

    # generate number of bins based on named algorithm
    if isinstance(bins, str):
        try:
            width = _hist_bin_selectors[bins](sx, (bottom, top))
        except IndexError:
            width = 1
        if numpy.isnan(width):
            width = 1
        try:
            if width:
                bins = int(numpy.ceil((top - bottom) / width))
            else:
                bins = 1
        except OverflowError:
            bins = 10
        # The spark graphs get hard to read if the bin slices are too thin, so we will max out at 50 bins
        if bins > 50:
            bins = 50
        # The spark graphs look weird if the bin slices are too fat, so we will min at 10 bins
        if bins < 10:
            bins = 10

    # generate actual bins based on integer number of bins
    if isinstance(bins, int):
        try:
            bins = numpy.linspace(
                bottom, top, num=bins, endpoint=True, retstep=False, dtype=None
            )
        except UnboundLocalError:
            raise

    return (
        bins,
        bottom,
        top,
        zeros_are_dropped,
        range_is_truncated,
        sx,
        pct_under,
        pct_over,
    )


def _compress_by_mask(arr, mask=None):
    if arr is None:
        return None
    if isinstance(arr, numpy.ma.core.MaskedArray):
        if mask is None:
            mask = arr
        arr = arr._data
    if isinstance(arr, pandas.Series):
        arr = arr.data
    if isinstance(arr, memoryview):
        arr = numpy.asarray(arr)
    data = numpy.ndarray.ravel(arr)
    if mask._mask is not _nomask:
        data = data.compress(numpy.logical_not(numpy.ndarray.ravel(mask._mask)))
    return data


def histogram_data(
    data_to_bin,
    bins="sturges",
    weights=None,
    ch_weights=None,
    density=None,
    percentiles_to_cut=1,
    maxrange=None,
    fixedrange=None,
    threshold_for_dropping_zeros=0.35,
):
    (
        bins,
        bottom,
        top,
        zeros_are_dropped,
        range_is_truncated,
        sx,
        pct_under,
        pct_over,
    ) = get_histogram_bins(
        data_to_bin,
        bins=bins,
        percentiles_to_cut=percentiles_to_cut,
        maxrange=maxrange,
        fixedrange=fixedrange,
        threshold_for_dropping_zeros=threshold_for_dropping_zeros,
    )

    h, bin_edges = numpy.histogram(
        _compress_by_mask(sx),
        bins=bins,
        range=(bottom, top),
        weights=_compress_by_mask(weights, sx),
        density=False,
    )
    if ch_weights is not None:
        ch_h, _ = numpy.histogram(
            _compress_by_mask(sx),
            bins=bin_edges,
            weights=_compress_by_mask(ch_weights, sx),
            density=False,
            range=(bottom, top),
        )
    else:
        ch_h = None

    if density:
        if fixedrange is not None and ch_weights is not None:
            # Get adjusted density for choice heights
            db = numpy.array(numpy.diff(bin_edges), float)
            if weights is None:
                h = h / db / sx._data.size
            else:
                h = h / db / weights.sum()
            ch_h = ch_h / db / ch_weights.sum()
        else:
            db = numpy.array(numpy.diff(bin_edges), float)
            h = h / db / h.sum()
            ch_h = ch_h / db / ch_h.sum()

    return (
        h,
        bin_edges,
        zeros_are_dropped,
        range_is_truncated,
        ch_h,
        pct_under,
        pct_over,
    )


_histogram_notes = {
    hexcolor(
        "orange"
    ): "Histograms are orange if the zeros are numerous and have been excluded.",
    hexcolor(
        "red"
    ): "Histograms are red if the zeros are numerous and have been excluded, "
    "and the displayed range truncates some extreme outliers.",
    hexcolor(
        "forest"
    ): "Histograms are green if the displayed range truncates some extreme outliers.",
    hexcolor("ocean"): None,
    hexcolor(
        "night"
    ): "Histograms are purple if the data is represented as discrete values.",
}


def draw_histogram_figure(
    bin_heights,
    bin_edges,
    zeros_are_dropped,
    range_is_truncated,
    title=None,
    xlabel=None,
    ylabel=None,
    xticks=False,
    yticks=False,
    xticklabels=None,
    yticklabels=False,
    frame_off="all",
    notetaker=None,
    bgcolor=None,
    figwidth=0.75,
    figheight=0.2,
    return_format="svg",
    attach_metadata=True,
    ch_heights=None,
    discrete=False,
    left_thermo=None,
    right_thermo=None,
):
    from .._optional import pyplot as plt

    fig, ax = plt.subplots(nrows=1, ncols=1, sharex=False, sharey=False)
    widths = bin_edges[1:] - bin_edges[:-1]
    if discrete:
        widths = widths * 0.8
        color = hexcolor("night")
    elif zeros_are_dropped:
        if range_is_truncated and not left_thermo and not right_thermo:
            color = hexcolor("red")
        else:
            color = hexcolor("orange")
    elif range_is_truncated and not left_thermo and not right_thermo:
        color = hexcolor("forest")
    else:
        color = hexcolor("ocean")
    # rects1 = ax.bar(
    #     bin_edges[:-1],
    #     bin_heights,
    #     widths,
    #     color=color,
    #     align="center" if discrete else "edge",
    # )

    chart_top = bin_heights.max()

    if ch_heights is not None:
        # rects2 = ax.bar(
        #     bin_edges[:-1],
        #     height=numpy.zeros_like(ch_heights),
        #     width=widths,
        #     bottom=ch_heights,
        #     edgecolor="black",
        #     linewidth=1.0,
        #     align="center" if discrete else "edge",
        #     clip_on=False,
        # )
        chart_top = max(chart_top, ch_heights.max())

    if left_thermo is not None:
        leftmost, rightmost = bin_edges[0], bin_edges[-1]
        full_width = rightmost - leftmost
        ax.bar(
            numpy.asarray(
                [
                    leftmost - full_width * 0.05,
                ]
            ),
            height=chart_top * left_thermo,
            width=numpy.asarray(
                [
                    full_width * 0.035,
                ]
            ),
            color="red",
            align="edge",
        )
        ax.bar(
            numpy.asarray(
                [
                    leftmost - full_width * 0.05,
                ]
            ),
            height=chart_top * (1 - left_thermo),
            width=numpy.asarray(
                [
                    full_width * 0.035,
                ]
            ),
            bottom=chart_top * left_thermo,
            color="#e9edd3",
            align="edge",
        )
        # Thermo-border
        ax.bar(
            numpy.asarray(
                [
                    leftmost - full_width * 0.05,
                ]
            ),
            height=chart_top,
            width=numpy.asarray(
                [
                    full_width * 0.035,
                ]
            ),
            color="#FFFFFF00",
            align="edge",
            edgecolor="black",
            linewidth=0.5,
            clip_on=False,
        )

    if right_thermo is not None:
        leftmost, rightmost = bin_edges[0], bin_edges[-1]
        full_width = rightmost - leftmost
        ax.bar(
            numpy.asarray(
                [
                    rightmost + full_width * 0.015,
                ]
            ),
            height=chart_top * right_thermo,
            width=numpy.asarray(
                [
                    full_width * 0.035,
                ]
            ),
            color="red",
            align="edge",
        )
        ax.bar(
            numpy.asarray(
                [
                    rightmost + full_width * 0.015,
                ]
            ),
            height=chart_top * (1 - right_thermo),
            width=numpy.asarray(
                [
                    full_width * 0.035,
                ]
            ),
            bottom=chart_top * right_thermo,
            color="#e9edd3",
            align="edge",
        )
        # Thermo-border
        ax.bar(
            numpy.asarray(
                [
                    rightmost + full_width * 0.015,
                ]
            ),
            height=chart_top,
            width=numpy.asarray(
                [
                    full_width * 0.035,
                ]
            ),
            color="#FFFFFF00",
            align="edge",
            edgecolor="black",
            linewidth=0.5,
            clip_on=False,
        )

    if figheight:
        fig.set_figheight(figheight)
    if figwidth:
        fig.set_figwidth(figwidth)
    fig.set_dpi(300)

    if frame_off == "all" or frame_off is True or frame_off == 1:
        frame_off = {"left", "right", "top", "bottom"}
    elif frame_off is None:
        frame_off = set()
    elif not isinstance(frame_off, set):
        print(type(frame_off))
        print(frame_off)
        frame_off = set(frame_off)

    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    if title:
        ax.set_title(title)
    if not isinstance(xticks, (list, tuple, numpy.ndarray)) and not xticks:
        ax.get_xaxis().set_ticks([])
    else:
        frame_off.discard("bottom")
        if xticklabels is not None:
            ax.get_xaxis().set_ticklabels(xticklabels)
            ax.get_xaxis().set_ticks(xticks)
    if not isinstance(yticks, (list, tuple, numpy.ndarray)) and not yticks:
        ax.get_yaxis().set_ticks([])
    else:
        frame_off.discard("left")
        if yticklabels is not None:
            ax.get_yaxis().set_ticklabels(yticklabels)
            ax.get_yaxis().set_ticks(yticks)

    if frame_off is not None:
        if isinstance(frame_off, str):
            frame_off = [frame_off]
        from .plotting import adjust_spines

        sides = {"left", "right", "top", "bottom"}
        for side in frame_off:
            sides.remove(side)
        adjust_spines(ax, sides)

    if bgcolor is not None:
        ax.set_axis_bgcolor(bgcolor)

    if notetaker is not None and color != hexcolor("ocean"):
        notetaker.add(_histogram_notes[color])

    if return_format.lower() == "svg":
        from .plotting import plot_as_svg_xhtml

        pad_inches = 0.1
        try:
            if "left" in frame_off and "bottom" in frame_off:
                ax.set_position([0, 0, 1, 1], which="both")
                ax.set_frame_on(False)
                pad_inches = 0
            elif figheight <= 1.0:
                plt.tight_layout(pad=figheight / 2)
                pad_inches = 0
            else:
                plt.tight_layout(pad=0.5)
        except ValueError:
            pass  # tight layout sometimes fails when the figsize is too small
        tooltip = _histogram_notes[color] or ""
        if ch_heights is not None:
            if len(tooltip):
                tooltip += "\n"
            tooltip += (
                "Black marks indicate frequencies considering only chosen alternatives."
            )
        if left_thermo or right_thermo:
            if len(tooltip):
                tooltip += "\n"
            tooltip += (
                "Side thermometers indicate the fraction of all observations "
                "at or beyond the limits, and are scaled independently."
            )

        ret = plot_as_svg_xhtml(plt, tooltip=tooltip, pad_inches=pad_inches)
        plt.clf()
        plt.close()  # do not showing empty windows?
        if attach_metadata:
            ret.metadata = dict(
                bin_heights=bin_heights,
                bin_edges=bin_edges,
                zeros_are_dropped=zeros_are_dropped,
                range_is_truncated=range_is_truncated,
            )
            ret.attrib["zeros_are_dropped"] = str(zeros_are_dropped)
            ret.attrib["range_is_truncated"] = str(range_is_truncated)
        ret.attrib["style"] = f"min-width:{figwidth}in"
        return ret


def histogram_figure(
    data_to_bin,
    bins="sturges",
    weights=None,
    ch_weights=None,
    return_all=False,
    percentiles_to_cut=1,
    maxrange=None,
    fixedrange=None,
    piecerange=None,
    mask=None,
    **kwargs,
):
    density = None if ch_weights is None else True
    discrete = bins == "discrete"

    if mask is not None:
        data_to_bin = data_to_bin[mask]
        if weights is not None:
            weights = weights[mask]
        if ch_weights is not None:
            ch_weights = ch_weights[mask]

    threshold_for_dropping_zeros = 0.35

    if piecerange is not None and piecerange != (None, None):
        if piecerange[0] is None:
            fixedrange = (None, piecerange[1])
        elif piecerange[1] is None:
            fixedrange = (0, data_to_bin.max() + 1e-8)
            threshold_for_dropping_zeros = 100
        else:
            fixedrange = (0, piecerange[1] - piecerange[0])
            threshold_for_dropping_zeros = 100
        percentiles_to_cut = 0
        discrete = False

    if discrete:
        bin_labels, bin_inv = numpy.unique(data_to_bin, return_inverse=True)
        bin_heights = numpy.bincount(bin_inv, weights)
        bin_edges = numpy.arange(len(bin_heights) + 1)
        zeros_are_dropped, range_is_truncated = False, False
        if ch_weights is not None:
            ch_heights = numpy.bincount(bin_inv, ch_weights)
        else:
            ch_heights = None
        if density:
            bin_heights = bin_heights / (bin_heights.sum())
            if ch_heights is not None:
                ch_heights = ch_heights / (ch_heights.sum())
        if (("xticks" not in kwargs) and (kwargs.get("figwidth", 1) >= 2)) or (
            kwargs.get("xticks") is True
        ):
            kwargs["xticklabels"] = bin_labels
            kwargs["xticks"] = bin_edges[:-1]
        pct_under, pct_over = 0, 0
    else:
        (
            bin_heights,
            bin_edges,
            zeros_are_dropped,
            range_is_truncated,
            ch_heights,
            pct_under,
            pct_over,
        ) = histogram_data(
            data_to_bin,
            bins,
            weights,
            ch_weights,
            density=density,
            percentiles_to_cut=percentiles_to_cut,
            maxrange=maxrange,
            fixedrange=fixedrange,
            threshold_for_dropping_zeros=threshold_for_dropping_zeros,
        )
        bin_labels = None

    if piecerange is not None and piecerange[0] is not None:
        bin_edges += piecerange[0]

    try:
        f = draw_histogram_figure(
            bin_heights,
            bin_edges,
            zeros_are_dropped,
            range_is_truncated,
            ch_heights=ch_heights,
            discrete=discrete,
            left_thermo=pct_under or None,
            right_thermo=pct_over or None,
            **kwargs,
        )
    except Exception:
        print("bin_heights\n", bin_heights)
        print("bin_edges\n", bin_edges)
        print("zeros_are_dropped\n", zeros_are_dropped)
        print("range_is_truncated\n", range_is_truncated)
        raise
    if return_all:
        return f, bin_heights, bin_edges, zeros_are_dropped, range_is_truncated
    else:
        return f


def sizable_histogram_figure(
    *args, sizer=1, discrete=None, on_error="ignore", **kwargs
):
    kwargs["figwidth"] = sizer
    kwargs["figheight"] = sizer / 3.5
    if sizer >= 2:
        kwargs["frame_off"] = {"top", "right"}
        kwargs["xticks"] = True
        # kwargs['yticks'] = True
    if sizer < 4:
        kwargs.pop("title", None)
    if sizer < 3:
        kwargs.pop("xlabel", None)
        kwargs.pop("ylabel", None)
    if discrete is not None:
        if discrete:
            kwargs["bins"] = "discrete"
    try:
        return histogram_figure(*args, **kwargs)
    except Exception:
        if on_error == "ignore":
            return None
        else:
            raise


def seems_like_discrete_data(arr, dictionary=None, return_uniques=False):
    if isinstance(arr.dtype, pandas.CategoricalDtype):
        # print('seems_like_discrete_data? YES categorical')
        if return_uniques:
            return (True, list(arr.cat.categories))
        return True
    elif numpy.issubdtype(arr.dtype, numpy.bool_):
        # print('seems_like_discrete_data? YES bool')
        if return_uniques:
            return (True, [False, True])
        return True
    else:
        pass
        # print('seems_like_discrete_data? not bool but',arr.dtype)
    if dictionary is None:
        if len(numpy.unique(arr[:100])) < 6:
            if len(numpy.unique(arr[:1000])) < 6:
                uniques = numpy.unique(arr)
                if len(uniques) < 6:
                    # print('seems_like_discrete_data? YES uniques < 6')
                    if return_uniques:
                        return (True, uniques)
                    return True
        # print('seems_like_discrete_data? too many and no dictionary')
    else:
        uniq = numpy.unique(arr)
        not_in_dict = 0
        for i in uniq:
            if i not in dictionary:
                not_in_dict += 1
        if not_in_dict > 2:
            # print(f'seems_like_discrete_data? dictionary but {not_in_dict} missing keys')
            if return_uniques:
                return False, None
            return False
        else:
            # print(f'seems_like_discrete_data? dictionary with {not_in_dict} missing keys')
            if return_uniques:
                return True, uniq
            return True
    if return_uniques:
        return False, None
    return False



import keyword as _keyword
import re
import re as _re
from collections.abc import Mapping, MutableMapping
from numbers import Number as _Number

import numpy as _numpy

from ..util.naming import parenthize

_ParameterRef_C_repr_txt = "P"
_DataRef_repr_txt = "X"

_null_ = "_"

_boolmatch = re.compile(r"^boolean\((.+)==(.+)\)$")


def _what_is(thing):
    if isinstance(thing, (ParameterRef, DataRef)):
        return repr(thing)
    if isinstance(thing, (str, int, float)):
        return f"{thing.__class__.__name__}({thing})"
    if isinstance(thing, LinearComponent):
        return f"{thing.__class__.__name__}({thing!r})"
    return f"<{thing.__class__.__name__}>"


def _unsupported_operands(op, a, b):
    return f"unsupported operands for {op}: '{_what_is(a)}' and '{_what_is(b)}'"


class UnicodeRef(str):
    """
    A common base class for all larch named reference types.

    This class itself has no features and should not be instantiated.
    Instead create :class:`ParameterRef` or :class:`DataRef` objects as needed.
    """


class Ref_Gen:
    def __init__(self, kind):
        self._kind = kind

    def __getattr__(self, key):
        return self._kind(key)

    def __call__(self, arg):
        return self._kind(str(arg))

    def __getitem__(self, arg):
        return self._kind(str(arg))


class ParameterRef(UnicodeRef):
    _precedence = 99

    def __init__(self, *args):
        self._formatting = None

    def set_fmt(self, formatting):
        self._formatting = formatting
        return self

    def __repr__(self):
        if _re.match("[_A-Za-z][_a-zA-Z0-9]*$", self) and not _keyword.iskeyword(self):
            return f"{_ParameterRef_C_repr_txt}.{self}"
        else:
            return f"{_ParameterRef_C_repr_txt}('{self}')"

    def __eq__(self, other):
        if isinstance(other, str) and not isinstance(other, DataRef):
            if str(self) == str(other):
                return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def __pos__(self):
        return self

    def __add__(self, other):
        if isinstance(self, ParameterRef):
            if other == 0:
                return self
            if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
                return LinearComponent(param=str(self), data="1") + other
            if isinstance(other, _Number):
                from .linear_math import ParameterAdd

                return ParameterAdd(self, other)
        elif isinstance(other, ParameterRef):
            if self == 0:
                return other
            if isinstance(self, (LinearComponent, LinearFunction)):
                return self + LinearComponent(param=str(other), data="1")
            if isinstance(self, _Number):
                from .linear_math import ParameterAdd

                return ParameterAdd(self, other)
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} + {_what_is(other)}")

    def __sub__(self, other):
        if isinstance(self, ParameterRef):
            if other == 0:
                return self
            if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
                return LinearComponent(param=str(self), data="1") - other
            if isinstance(other, _Number):
                from .linear_math import ParameterSubtract

                return ParameterSubtract(self, other)
        elif isinstance(other, ParameterRef):
            if self == 0:
                return other
            if isinstance(self, (LinearComponent, LinearFunction)):
                return self - LinearComponent(param=str(other), data="1")
            if isinstance(self, _Number):
                from .linear_math import ParameterSubtract

                return ParameterSubtract(self, other)
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} - {_what_is(other)}")

    def __mul__(self, other):
        if isinstance(self, ParameterRef):
            if isinstance(other, DataRef):
                return LinearComponent(param=str(self), data=str(other))
            if isinstance(other, _Number):
                #  return LinearComponent(param=str(self), data=str(other))
                return LinearComponent(param=str(self), data="1", scale=other)
            if isinstance(other, ParameterRef):
                if self == _null_:
                    return other
                if other == _null_:
                    return self
                from .linear_math import ParameterMultiply

                return ParameterMultiply(self, other)
            if isinstance(other, LinearComponent):
                if self == _null_:
                    return other
                if other.param == _null_:
                    return LinearComponent(
                        param=str(self), data=str(other.data), scale=other.scale
                    )
                from .linear_math import ParameterMultiply

                return ParameterMultiply(self, other)
        elif isinstance(other, ParameterRef):
            if isinstance(self, DataRef):
                return LinearComponent(param=str(other), data=str(self))
            if isinstance(self, _Number):
                # return LinearComponent(param=str(other), data=str(self))
                return LinearComponent(param=str(other), data="1", scale=self)
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} * {_what_is(other)}")

    def __truediv__(self, other):
        if isinstance(self, ParameterRef):
            if isinstance(other, ParameterRef):
                from .linear_math import ParameterDivide

                return ParameterDivide(self, other)
            elif isinstance(other, _Number):
                return LinearComponent(param=str(self), data="1", scale=1 / other)
            elif isinstance(other, DataRef):
                return LinearComponent(param=str(self), data=str(1 / other), scale=1)
        elif isinstance(other, ParameterRef):
            return NotImplemented
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} / {_what_is(other)}")

    def value(self, *args):
        """
        Get the value of the parameter in a given model.

        Parameters
        ----------
        m : Model
            The model from which to extract a parameter value.

        Returns
        -------
        float
        """
        s = str(self)
        for m in args:
            if isinstance(m, dict) and s in m:
                return m[s]
            else:
                try:
                    return m.get_value(s)
                except KeyError:
                    pass
        raise KeyError(s)

    def string(self, m):
        """
        Get the value of the parameter in a given model, as a formatted string.

        Parameters
        ----------
        m : Model
            The model from which to extract a parameter value.

        Returns
        -------
        str
        """
        if self._formatting is None:
            return f"{self.value(m):.3g}"
        else:
            return self._formatting.format(self.value(m))

    def valid(self, m):
        """
        Check if this ParameterRef would give a value for a given model.

        Parameters
        ----------
        m : Model
            The model from which to extract a parameter value.

        Returns
        -------
        bool
                False if the value method would raise an exception, and True otherwise.
        """
        if str(self) in m:
            return True
        return False

    def as_pmath(self):
        from .linear_math import ParameterNoop

        return ParameterNoop(self)

    def __xml__(self, resolve_parameters=None, value_in_tooltips=True):
        if resolve_parameters is not None:
            if value_in_tooltips:
                p_display = repr(self)
                p_tooltip = self.string(resolve_parameters)
            else:
                p_display = self.string(resolve_parameters)
                p_tooltip = repr(self)
        else:
            p_display = repr(self)
            p_tooltip = "This is a Parameter"

        from xmle import Elem

        x = Elem("div")
        if use_tooltips:
            a_p = x.elem("div", attrib={"class": "tooltipped"}, text=p_display)
            a_p.elem("span", attrib={"class": "tooltiptext"}, text=p_tooltip)
        else:
            a_p = x.elem("span", attrib={"class": "Larch_Parameter"}, text=p_display)
        return x


class DataRef(UnicodeRef):
    def __repr__(self):
        if _re.match("[_A-Za-z][_a-zA-Z0-9]*$", self) and not _keyword.iskeyword(self):
            return f"{_DataRef_repr_txt}.{self}"
        else:
            return f"{_DataRef_repr_txt}('{self}')"

    def __eq__(self, other):
        if isinstance(other, str) and not isinstance(other, ParameterRef):
            if str(self) == str(other):
                return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def __pos__(self):
        return self

    def __add__(self, other):
        if isinstance(self, (DataRef, _Number)) and isinstance(
            other, (DataRef, _Number)
        ):
            if self == "0" or self == "0.0" or self == 0:
                return other
            if other == "0" or other == "0.0" or other == 0:
                return self
            # Double zero is trapped here as it is used to flag duplicate terms in a utility function.
            if self == "00":
                return DataRef(f"0+{parenthize(self)}")
            if other == "00":
                return DataRef(f"{parenthize(self)}+0")
            return DataRef(f"{parenthize(self)}+{parenthize(other, True)}")
        if isinstance(self, DataRef) and isinstance(other, ParameterRef):
            return P(_null_) * self + other

        # Don't return NotImplemented just raise TypeError when adding a DataRef and a plain string.
        # This will disallow the __radd__ method on the plain string.
        if (
            isinstance(self, (DataRef, _Number))
            and isinstance(other, str)
            and not isinstance(other, DataRef)
        ):
            raise TypeError(_unsupported_operands("-", self, other))
        if (
            isinstance(other, (DataRef, _Number))
            and isinstance(self, str)
            and not isinstance(self, DataRef)
        ):
            raise TypeError(_unsupported_operands("-", self, other))

        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} + {_what_is(other)}")

    def __sub__(self, other):
        if isinstance(self, (DataRef, _Number)) and isinstance(
            other, (DataRef, _Number)
        ):
            return DataRef(f"{parenthize(self)}-{parenthize(other, True)}")
        if isinstance(self, DataRef) and isinstance(other, ParameterRef):
            return P(_null_) * self - other
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} - {_what_is(other)}")

    def __mul__(self, other):
        if isinstance(self, DataRef):
            if isinstance(other, (DataRef, _Number)):
                if self == "1" or self == "1.0" or self == 1:
                    return other
                if other == "1" or other == "1.0" or other == 1:
                    return self
                if isinstance(other, _Number):
                    return P(_null_) * other * self
                if self == other and self[:8] == "boolean(" and self[-1:] == ")":
                    # Squaring a boolean does not change it
                    return self
                if (
                    self[:8] == "boolean("
                    and self[-1:] == ")"
                    and other[:8] == "boolean("
                    and other[-1:] == ")"
                ):
                    # Check for two mutually exclusive conditions
                    boolmatch1 = _boolmatch.match(self)
                    if boolmatch1:
                        boolmatch2 = _boolmatch.match(other)
                        if boolmatch2:
                            if boolmatch1.group(1) == boolmatch2.group(1):
                                if boolmatch1.group(2) != boolmatch2.group(2):
                                    return DataRef("0")
                return DataRef(f"{parenthize(self)}*{parenthize(other, True)}")
            if isinstance(other, ParameterRef):
                return LinearComponent(param=str(other), data=str(self))
            if isinstance(other, LinearComponent):
                return LinearComponent(
                    param=str(other.param),
                    data=str(self * other.data),
                    scale=other.scale,
                )
            if isinstance(other, LinearFunction):
                return LinearFunction([self * c for c in other])
        elif isinstance(other, DataRef):
            if isinstance(self, (DataRef, _Number)):
                if self == "1" or self == "1.0" or self == 1:
                    return other
                if other == "1" or other == "1.0" or other == 1:
                    return self
                if isinstance(self, _Number):
                    return P(_null_) * self * other
                return DataRef(f"{parenthize(self)}*{parenthize(other, True)}")
            if isinstance(self, ParameterRef):
                return LinearComponent(param=str(self), data=str(other))
            if isinstance(self, LinearComponent):
                return LinearComponent(
                    param=str(self.param), data=str(self.data * other)
                )
            if isinstance(self, LinearFunction):
                return LinearFunction([c * other for c in self])
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} * {_what_is(other)}")

    def __truediv__(self, other):
        if isinstance(self, (DataRef, _Number)) and isinstance(
            other, (DataRef, _Number)
        ):
            return DataRef(f"{parenthize(self)}/{parenthize(other, True)}")
        if isinstance(self, ParameterRef) and isinstance(other, (DataRef, _Number)):
            return self * DataRef(f"1/{parenthize(other, True)}")
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} / {_what_is(other)}")

    def __and__(self, other):
        if isinstance(self, (DataRef, _Number)) and isinstance(
            other, (DataRef, _Number)
        ):
            return DataRef(f"{parenthize(self)}&{parenthize(other, True)}")
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} & {_what_is(other)}")

    def __or__(self, other):
        if isinstance(self, (DataRef, _Number)) and isinstance(
            other, (DataRef, _Number)
        ):
            return DataRef(f"{parenthize(self)}|{parenthize(other, True)}")
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} | {_what_is(other)}")

    def __xor__(self, other):
        if isinstance(self, (DataRef, _Number)) and isinstance(
            other, (DataRef, _Number)
        ):
            return DataRef(f"{parenthize(self)}^{parenthize(other, True)}")
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} ^ {_what_is(other)}")

    def __floordiv__(self, other):
        if isinstance(self, (DataRef, _Number)) and isinstance(
            other, (DataRef, _Number)
        ):
            return DataRef(f"{parenthize(self)}//{parenthize(other, True)}")
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} // {_what_is(other)}")

    def __pow__(self, other, modulo):
        if modulo is not None:
            raise NotImplementedError(
                f"no pow with modulo on {self.__class__.__name__}"
            )
        if isinstance(self, (DataRef, _Number)) and isinstance(
            other, (DataRef, _Number)
        ):
            return DataRef(f"{parenthize(self)}**{parenthize(other, True)}")
        return NotImplemented  # raise NotImplementedError(f"{_what_is(self)} ** {_what_is(other)}")

    def __invert__(self):
        return DataRef(f"~{parenthize(self, True)}")

    def __neg__(self):
        return DataRef(f"-{parenthize(self, True)}")

    def eval(self, namespace=None, *, globals=None, **more_namespace):
        import numpy

        from ..util.common_functions import hard_sigmoid, piece

        use_namespace = {
            "exp": numpy.exp,
            "log": numpy.log,
            "log1p": numpy.log1p,
            "fabs": numpy.fabs,
            "sqrt": numpy.sqrt,
            "absolute": numpy.absolute,
            "isnan": numpy.isnan,
            "isfinite": numpy.isfinite,
            "logaddexp": numpy.logaddexp,
            "fmin": numpy.fmin,
            "fmax": numpy.fmax,
            "nan_to_num": numpy.nan_to_num,
            "piece": piece,
            "hard_sigmoid": hard_sigmoid,
        }
        if namespace is not None:
            use_namespace.update(namespace)
        use_namespace.update(more_namespace)
        return eval(self, globals, use_namespace)


use_tooltips = False


class LinearComponent:
    def __init__(self, param: str, data: str = "1", scale: float = 1):
        self._param = param
        self._data = data
        self._scale = scale

    @property
    def param(self):
        return ParameterRef(self._param)

    @property
    def data(self):
        return DataRef(self._data)

    @property
    def scale(self):
        return self._scale

    def to_dict(self):
        out = {
            "param": str(self._param),
            "data": str(self._data),
        }
        if self.scale != 1:
            out["scale"] = self.scale
        return out

    def __pos__(self):
        return self

    def __neg__(self):
        return self.__class__(param=self._param, data=self._data, scale=-self._scale)

    def __repr__(self):
        try:
            if self.scale == 1.0:
                try:
                    data_is_1 = float(self.data) == 1
                except Exception:
                    data_is_1 = False
                if data_is_1:
                    return f"{self.param!r}"
                else:
                    return f"{self.param!r} * {self.data!r}"
            return f"{self.param!r} * {self.scale} * {self.data!r}"
        except AttributeError:
            return f"<{self.__class__.__name__} {id(self)} with error>"

    def _str_exponentiate(self):
        try:
            if self.scale == 1.0:
                try:
                    data_is_1 = float(self.data) == 1
                except Exception:
                    data_is_1 = False
                if data_is_1:
                    return f"exp({self.param!r})"
                else:
                    return f"exp({self.param!r}) * {self.data!r}"
            return f"exp({self.param!r}) * {self.scale} * {self.data!r}"
        except AttributeError:
            return f"<{self.__class__.__name__} {id(self)} with error>"

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            raise TypeError(
                f"unsupported operand type(s) for +: {type(other)} and {type(self)}"
            )

    def __add__(self, other):
        if isinstance(self, LinearComponent):
            if other == () or other == 0:
                return self
            elif isinstance(other, LinearComponent):
                if other.data == "0" or other.data == "0.0" or other.scale == 0:
                    return self
                if self.data == "0" or self.data == "0.0" or self.scale == 0:
                    return other
                return LinearFunction([self, other])
            elif isinstance(other, LinearFunction):
                return LinearFunction([self, *other])
            elif isinstance(other, ParameterRef):
                return self + LinearComponent(param=str(other))
            elif isinstance(other, DataRef):
                return self + LinearComponent(param=_null_, data=str(other))
            else:
                try:
                    return self.as_pmath() + other
                except NotImplementedError:
                    pass
        elif isinstance(other, LinearComponent):
            if self == () or self == 0:
                return other
            elif isinstance(self, ParameterRef):
                # return LinearFunction([LinearComponent(param=str(self)), other])
                return LinearComponent(param=str(self)) + other
            elif isinstance(self, DataRef):
                return LinearComponent(param=_null_, data=str(self)) + other
            elif isinstance(self, LinearFunction):
                return (
                    LinearFunction(
                        [
                            *self,
                        ]
                    )
                    + other
                )
            else:
                try:
                    return other + self.as_pmath()
                except NotImplementedError:
                    pass
        raise NotImplementedError(f"{_what_is(self)} + {_what_is(other)}")

    def __sub__(self, other):
        if isinstance(self, LinearComponent):
            if other == () or other == 0:
                return self
            elif isinstance(other, LinearComponent):
                return LinearFunction([self, -other])
            elif isinstance(other, LinearFunction):
                return LinearFunction([self, *(-other)])
            elif isinstance(other, ParameterRef):
                return LinearFunction([self, -LinearComponent(param=str(other))])
            elif isinstance(other, ParameterRef):
                return LinearFunction(
                    [self, -LinearComponent(param=_null_, data=str(other))]
                )
            else:
                try:
                    return self.as_pmath() - other
                except NotImplementedError:
                    pass
        raise NotImplementedError(f"{_what_is(self)} + {_what_is(other)}")

    def __mul__(self, other):
        if isinstance(self, LinearComponent):
            if isinstance(
                other,
                (
                    int,
                    float,
                ),
            ):
                return self.__class__(
                    param=str(self.param),
                    data=str(self.data),
                    scale=self.scale * other,
                )
            if isinstance(other, (DataRef,)):
                return self.__class__(
                    param=str(self.param),
                    data=str(self.data * other),
                    scale=self.scale,
                )
            if isinstance(other, (LinearComponent,)):
                if self.param == _null_:
                    return self.__class__(
                        param=str(other.param),
                        data=str(self.data * other.data),
                        scale=self.scale * other.scale,
                    )
                if other.param == _null_:
                    return self.__class__(
                        param=str(self.param),
                        data=str(self.data * other.data),
                        scale=self.scale * other.scale,
                    )
                from .linear_math import ParameterMultiply

                return ParameterMultiply(
                    self.as_pmath(),
                    other.as_pmath(),
                )
            if isinstance(other, ParameterRef):
                if other == _null_:
                    return self
                elif self.param == _null_:
                    return self.__class__(
                        param=str(other),
                        data=str(self.data),
                        scale=self.scale,
                    )
            try:
                return self.as_pmath() * other
            except NotImplementedError:
                pass
        if isinstance(other, LinearComponent) and isinstance(self, _Number):
            return other.__class__(
                param=str(other.param),
                data=str(other.data),
                scale=other.scale * self,
            )
        raise NotImplementedError(f"{_what_is(self)} * {_what_is(other)}")

    def __truediv__(self, other):
        if isinstance(self, LinearComponent):
            if isinstance(
                other,
                (
                    int,
                    float,
                ),
            ):
                return self.__class__(
                    param=str(self.param),
                    data=str(self.data),
                    scale=self.scale / other,
                )
            elif isinstance(other, (DataRef,)):
                return self.__class__(
                    param=str(self.param),
                    data=str(self.data / other),
                    scale=self.scale,
                )
            elif isinstance(other, (LinearComponent,)):
                from .linear_math import ParameterDivide

                return ParameterDivide(
                    self.as_pmath(),
                    other.as_pmath(),
                )
            else:
                try:
                    return self.as_pmath() / other
                except NotImplementedError:
                    pass
        raise NotImplementedError(f"{_what_is(self)} / {_what_is(other)}")

    def __iter__(self):
        return iter(LinearFunction([self]))

    def __eq__(self, other):
        if isinstance(other, LinearFunction) and len(other) == 1:
            other = other[0]
        if not isinstance(other, LinearComponent):
            return False
        if self.param != other.param:
            return False
        if self.data != other.data:
            return False
        if self.scale != other.scale:
            return False
        return True

    def __xml__(
        self,
        exponentiate_parameter=False,
        resolve_parameters=None,
        value_in_tooltips=True,
    ):
        from xmle import Elem

        if use_tooltips:
            x = Elem("div")
            # x << tooltipped_style()
            if resolve_parameters is not None:
                if exponentiate_parameter:
                    if value_in_tooltips:
                        p_tooltip = (
                            f"exp({self.param.string(resolve_parameters)}) "
                            f"= {self.param.value(resolve_parameters):.4g}"
                        )
                        p_display = f"{repr(self.param)}"
                    else:
                        p_display = f"{self.param.string(resolve_parameters)}"
                        p_tooltip = f"exp({repr(self.param)})"
                else:
                    if value_in_tooltips:
                        p_tooltip = self.param.string(resolve_parameters)
                        p_display = repr(self.param)
                    else:
                        p_display = self.param.string(resolve_parameters)
                        p_tooltip = repr(self.param)
            else:
                p_display = repr(self.param)
                p_tooltip = "This is a Parameter"

            data_tail = " * "
            try:
                if float(self.data) == 1:
                    data_tail = ""
            except Exception:
                pass

            if self.scale == 1.0:
                if exponentiate_parameter:
                    x.elem("span", tail="exp(")
                    a_p = x.elem(
                        "div",
                        attrib={"class": "tooltipped"},
                        text=p_display,
                        tail=")" + data_tail,
                    )
                    a_p.elem("span", attrib={"class": "tooltiptext"}, text=p_tooltip)
                else:
                    a_p = x.elem(
                        "div",
                        attrib={"class": "tooltipped"},
                        text=p_display,
                        tail=data_tail,
                    )
                    a_p.elem("span", attrib={"class": "tooltiptext"}, text=p_tooltip)
            else:
                if exponentiate_parameter:
                    x.elem("span", tail="exp(")
                    a_p = x.elem(
                        "div",
                        attrib={"class": "tooltipped"},
                        text=p_display,
                        tail=f" * {self.scale}){data_tail}",
                    )
                    a_p.elem("span", attrib={"class": "tooltiptext"}, text=p_tooltip)
                else:
                    a_p = x.elem(
                        "div",
                        attrib={"class": "tooltipped"},
                        text=p_display,
                        tail=f" * {self.scale}{data_tail}",
                    )
                    a_p.elem("span", attrib={"class": "tooltiptext"}, text=p_tooltip)
            if data_tail == " * ":
                a_x = x.elem(
                    "div", attrib={"class": "tooltipped"}, text=repr(self.data)
                )
                a_x.elem("span", attrib={"class": "tooltiptext"}, text="This is Data")

        else:
            x = Elem("pre")
            # x << tooltipped_style()
            if resolve_parameters is not None:
                if exponentiate_parameter:
                    if value_in_tooltips:
                        p_display = f"{repr(self.param)}"
                    else:
                        p_display = f"{self.param.string(resolve_parameters)}"
                else:
                    if value_in_tooltips:
                        p_display = repr(self.param)
                    else:
                        p_display = self.param.string(resolve_parameters)
            else:
                p_display = repr(self.param)

            data_tail = " * "
            try:
                if float(self.data) == 1:
                    data_tail = ""
            except Exception:
                pass

            if self.scale == 1.0:
                if exponentiate_parameter:
                    x.elem("span", tail="exp(")
                    a_p = x.elem(
                        "span",
                        attrib={"class": "LinearFunc_Param"},
                        text=p_display,
                        tail=")" + data_tail,
                    )
                else:
                    a_p = x.elem(
                        "span",
                        attrib={"class": "LinearFunc_Param"},
                        text=p_display,
                        tail=data_tail,
                    )
            else:
                if exponentiate_parameter:
                    x.elem("span", tail="exp(")
                    a_p = x.elem(
                        "span",
                        attrib={"class": "LinearFunc_Param"},
                        text=p_display,
                        tail=f" * {self.scale}){data_tail}",
                    )
                else:
                    a_p = x.elem(
                        "span",
                        attrib={"class": "LinearFunc_Param"},
                        text=p_display,
                        tail=f" * {self.scale}{data_tail}",
                    )
            if data_tail == " * ":
                a_x = x.elem(
                    "span", attrib={"class": "LinearFunc_Data"}, text=repr(self.data)
                )
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()

    def evaluate(self, p_getter, x_namespace=None, exp_params=False, **kwargs):
        if self.data in x_namespace:
            x = x_namespace[self.data]
        elif self.data in kwargs:
            x = kwargs[self.data]
        else:
            x = self.data.eval(namespace=x_namespace, **kwargs)
        if exp_params:
            return self.scale * _numpy.exp(p_getter(str(self.param))) * x
        else:
            return self.scale * p_getter(str(self.param)) * x

    def __copy__(self):
        return self.__class__(
            param=str(self.param),
            data=str(self._data),
            scale=self.scale,
        )

    def as_pmath(self):
        from .linear_math import ParameterMultiply, ParameterNoop

        if self.data != "1":
            try:
                scale = float(self.data) * self.scale
            except Exception:
                pass
            else:
                if scale == 1:
                    return ParameterNoop(self.param)
                else:
                    return ParameterMultiply(self.param, scale)
            raise NotImplementedError("data is not 1")
        if self.scale == 1:
            return ParameterNoop(self.param)
        else:
            return ParameterMultiply(self.param, self.scale)


def _try_mangle(instance):
    try:
        instance.mangle()
    except AttributeError:
        pass  # print(f"No Mangle L: {err}")


def _try_mangle_h(instance_holder):
    try:
        instance_holder._instance.mangle()
    except AttributeError:
        pass  # print(f"No Mangle L2: {err}")


class LinearFunction:
    _instance = None

    def __init__(self, init=None):
        self._func = list()
        if init is not None and init != 0:
            # Copy contents of init, stabilizes functionality of iadd
            init_ = list(init)
            for i in init_:
                if isinstance(i, Mapping):
                    i = LinearComponent(**i)
                if isinstance(i, LinearComponent):
                    self._func.append(i)
                else:
                    raise TypeError(
                        f"members of {self.__class__.__name__} must be LinearComponent"
                    )

    def set_instance(self, instance):
        self._instance = instance

    def __fresh(self, instance):
        newself = LinearFunction()
        newself._instance = instance
        setattr(instance, self.private_name, newself)
        return newself

    def __get__(self, instance, owner):
        # LinearFunction newself

        if instance is None:
            return self
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        return newself

    def __set__(self, instance, values):
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        newself.__init__(values)
        # _try_mangle_h(newself)
        try:
            newself._instance.mangle()
        except AttributeError:
            pass  # print(f"No Mangle L2: {err}")
        else:
            pass  # print(f"Yes Mangle L2: {newself._instance}")

    def __delete__(self, instance):
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        newself.__init__()
        newself._instance = instance
        # _try_mangle_h(newself)
        try:
            newself._instance.mangle()
        except AttributeError:
            pass  # print(f"No Mangle L2: {err}")
        else:
            pass  # print(f"Yes Mangle L2: {newself._instance}")

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name

    def __getitem__(self, item):
        return self._func[item]

    def __setitem__(self, key, value: LinearComponent):
        self._func[key] = value
        _try_mangle(self._instance)

    def __delitem__(self, key):
        del self._func[key]
        _try_mangle(self._instance)

    def remove_data(self, data):
        """
        Remove all terms from this linear function with the given data.

        This operation mutates this linear function in-place.

        Parameters
        ----------
        data : str
                Name of data to remove, which must match exactly.

        Returns
        -------
        self : LinearFunction
        """
        i = len(self._func)
        while i > 0:
            i -= 1
            if self._func[i].data == data:
                del self._func[i]
        return self

    def remove_param(self, param):
        """
        Remove all terms from this linear function with the given parameter.

        This operation mutates this linear function in-place.

        Parameters
        ----------
        param : str
                Name of parameter to remove

        Returns
        -------
        self : LinearFunction
        """
        i = len(self._func)
        while i > 0:
            i -= 1
            if self._func[i].param == param:
                del self._func[i]
        return self

    def __len__(self):
        return len(self._func)

    def insert(self, index, value: LinearComponent):
        self._func.insert(index, value)
        _try_mangle(self._instance)

    def append(self, value: LinearComponent):
        self._func.append(value)
        _try_mangle(self._instance)

    def extend(self, values):
        for v in values:
            if not isinstance(v, LinearComponent):
                raise TypeError(f"cannot add type {type(v)} to LinearFunction")
        self._func.extend(values)
        _try_mangle(self._instance)

    def __add__(self, other):
        if isinstance(self, LinearFunction):
            if other == () or other == 0:
                return self
            if isinstance(other, LinearFunction):
                return self.__class__([*list(self), *list(other)])
            if isinstance(other, ParameterRef):
                other = LinearComponent(param=str(other))
            if isinstance(other, LinearComponent):
                result = self.__class__(self)
                if not (other.data == "0" or other.data == "0.0" or other.scale == 0):
                    result.append(other)
                return result
            from .linear_math import ParameterAdd, _ParameterOp

            if isinstance(other, (_ParameterOp, _Number)):
                try:
                    return ParameterAdd(self.as_pmath(), other)
                except NotImplementedError:
                    pass
        if isinstance(other, LinearFunction):
            if self == () or self == 0:
                return other
            from .linear_math import ParameterAdd, _ParameterOp

            if isinstance(self, (_ParameterOp, _Number)):
                try:
                    return ParameterAdd(self, other.as_pmath())
                except NotImplementedError:
                    pass
        raise NotImplementedError(f"{_what_is(self)} + {_what_is(other)}")

    def __iadd__(self, other):
        if isinstance(other, ParameterRef):
            other = LinearComponent(param=str(other))
        if other == () or other == 0:
            return self
        elif isinstance(other, LinearFunction):
            self._func.extend(other)
            _try_mangle(self._instance)
        elif isinstance(other, LinearComponent):
            self.append(other)
        else:
            raise TypeError(f"cannot add type {type(other)} to LinearFunction")
        return self

    def __pos__(self):
        return self

    def __neg__(self):
        return self.__class__(-i for i in self)

    def __sub__(self, other):
        if isinstance(self, LinearFunction):
            if other == () or other == 0:
                return self
            if isinstance(other, LinearFunction):
                return self.__class__([*list(self), *list(-other)])
            if isinstance(other, ParameterRef):
                other = LinearComponent(param=str(other))
            if isinstance(other, LinearComponent):
                result = self.__class__(self)
                result.append(-other)
                return result
            from .linear_math import ParameterSubtract, _ParameterOp

            if isinstance(other, (_ParameterOp, _Number)):
                try:
                    return ParameterSubtract(self.as_pmath(), other)
                except NotImplementedError:
                    pass
        if isinstance(other, LinearFunction):
            from .linear_math import ParameterSubtract, _ParameterOp

            if isinstance(self, (_ParameterOp, _Number)):
                try:
                    return ParameterSubtract(self, other.as_pmath())
                except NotImplementedError:
                    pass
        raise NotImplementedError(f"{_what_is(self)} + {_what_is(other)}")

    def __mul__(self, other):
        from .linear_math import ParameterMultiply, _ParameterOp

        if isinstance(self, LinearFunction) and isinstance(other, _Number):
            trial = LinearFunction()
            for component in self:
                trial.append(component * other)
            return trial
        if isinstance(self, LinearFunction) and isinstance(
            other, (ParameterRef, _ParameterOp)
        ):
            if isinstance(other, ParameterRef):
                if other == _null_:
                    return self
                trial = LinearFunction()
                for component in self:
                    if component.param == _null_:
                        trial.append(component * other)
                    else:
                        trial = None
                        break
                if trial is not None:
                    return trial
            return ParameterMultiply(self.as_pmath(), other)
        if isinstance(self, LinearFunction) and isinstance(other, LinearFunction):
            return sum(i * j for i in self for j in other)
        if isinstance(other, LinearFunction) and isinstance(self, _Number):
            trial = LinearFunction()
            for component in other:
                trial.append(self * component)
            return trial
        if isinstance(other, LinearFunction) and isinstance(
            self, (ParameterRef, _ParameterOp)
        ):
            if isinstance(self, ParameterRef):
                if self == _null_:
                    return other
                trial = LinearFunction()
                for component in other:
                    if component.param == _null_:
                        trial.append(self * component)
                    else:
                        trial = None
                        break
                if trial is not None:
                    return trial
            return ParameterMultiply(self, other.as_pmath())
        try:
            trial = LinearFunction()
            for component in self:
                trial.append(component * other)
            return trial
        except NotImplementedError:
            return NotImplemented

    def __truediv__(self, other):
        from .linear_math import ParameterDivide, _ParameterOp

        if isinstance(self, LinearFunction) and isinstance(
            other, (ParameterRef, _ParameterOp, _Number)
        ):
            return ParameterDivide(self.as_pmath(), other)
        if isinstance(other, LinearFunction) and isinstance(
            self, (ParameterRef, _ParameterOp, _Number)
        ):
            return ParameterDivide(self, other.as_pmath())
        return NotImplemented

    def __contains__(self, val):
        if isinstance(val, ParameterRef):
            for i in self:
                if i.param == val:
                    return True
            return False
        if isinstance(val, DataRef):
            for i in self:
                if i.data == val:
                    return True
            return False
        raise TypeError(
            "the searched for content must be of type ParameterRef or DataRef"
        )

    def _index_of(self, val):
        if isinstance(val, ParameterRef):
            for n, i in enumerate(self):
                if i.param == val:
                    return n
            raise KeyError("ParameterRef not found")
        if isinstance(val, DataRef):
            for n, i in enumerate(self):
                if i.data == val:
                    return n
            raise KeyError("DataRef not found")
        raise TypeError(
            "the searched for content must be of type ParameterRef or DataRef"
        )

    def reformat_param(self, container=None, pattern=None, repl=None, **kwargs):
        """
        Transform all the parameters in the LinearFunction.

        Parameters
        ----------
        container : str
                A format string, into which the previous parameters are formatted.
                Use this to append things to the parameter names.
        pattern : str
        repl : str
                Passed to `re.sub` with each existing parameter as the base string
                to be searched.

        Examples
        --------
        >>> from larch.roles import P,X
        >>> f = P.InVehTime * X.IVTT + P.OutOfVehTime * X.OVTT
        >>> f1 = f.reformat_param('{}_Suffix')
        >>> str(f1)
        '(P.InVehTime_Suffix * X.IVTT) + (P.OutOfVehTime_Suffix * X.OVTT)'
        >>> f2 = f.reformat_param(pattern='(Veh)', repl='Vehicle')
        >>> str(f2)
        '(P.InVehicleTime * X.IVTT) + (P.OutOfVehicleTime * X.OVTT)'

        """
        import re

        r = self.__class__()
        for i in self:
            if pattern is None:
                param = i.param
            else:
                if repl is None:
                    raise TypeError("must give repl with pattern")
                param = re.sub(pattern, repl, i.param, **kwargs)
            if container is None:
                container = "{}"
            r += LinearComponent(
                data=str(i.data), param=container.format(param), scale=i.scale
            )
        return r

    def reformat_data(self, container=None, pattern=None, repl=None, **kwargs):
        """
        Transform all the data in the LinearFunction.

        Parameters
        ----------
        container : str
                A format string, into which the previous data strings are formatted.
                Use this to apply common global transforms to the data.
        pattern : str
        repl : str
                Passed to `re.sub` with each existing data string as the base string
                to be searched.
        """
        import re

        r = self.__class__()
        for i in self:
            if pattern is None:
                data = i.data
            else:
                if repl is None:
                    raise TypeError("must give repl with pattern")
                data = re.sub(pattern, repl, i.data, **kwargs)
            if container is None:
                container = "{}"
            r += LinearComponent(
                data=container.format(data), param=str(i.param), scale=i.scale
            )
        return r

    def __code__(self):
        return " + ".join(f"({repr(i)})" for i in self)

    def __eq__(self, other):
        if not isinstance(other, LinearFunction):
            return False
        if len(self) != len(other):
            return False
        for i, j in zip(self, other):
            if i != j:
                return False
        return True

    def __repr__(self):
        if len(self):
            result = " + ".join(repr(i) for i in self)
            if len(result) < 80:
                return result
            else:
                return "  " + result.replace(" + ", "\n+ ")
        return f"<Empty {self.__class__.__name__}>"

    def __xml__(
        self,
        linebreaks=True,
        lineprefix="",
        exponentiate_parameters=False,
        resolve_parameters=None,
        value_in_tooltips=True,
    ):
        from xmle import Elem

        x = Elem("div" if use_tooltips else "pre", attrib={"class": "LinearFunc"})
        for n, i in enumerate(self):
            ix_ = list(
                i.__xml__(
                    exponentiate_parameter=exponentiate_parameters,
                    resolve_parameters=resolve_parameters,
                    value_in_tooltips=value_in_tooltips,
                )
            )
            if linebreaks:
                if n > 0 or lineprefix:
                    ix_.insert(0, Elem("br", tail=lineprefix + " + "))
            else:
                if n < len(self) - 1:
                    if ix_[-1].tail is None:
                        ix_[-1].tail = " + "
                    else:
                        ix_[-1].tail += " + "
            for ii in ix_:
                x << ii
        if len(self) == 0:
            x << Elem("span", text=repr(self))
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()

    def data(self, cls=None):
        if cls is None:
            return [_.data for _ in self]
        else:
            return [cls(_.data) for _ in self]

    def evaluate(
        self, param_source, x_namespace=None, exp_params=False, **more_x_namespace
    ):
        """
        Evaluate the linear function in the context of some parameters and data.

        Typically all of the data given will be scalar values (to compute a
        scalar result) or a single data item will be a vector of possible
        values (to get a vector result).

        Parameters
        ----------
        param_source : Model-like
                The source of the current parameter values.
        x_namespace : dict, optional
                A namespace of data values.
        exp_params : bool, default False
                Whether to take the exponential of parameters (i.e. for
                a quantity function).
        **more_x_namespace : any
                More data values

        Returns
        -------
        numeric or array-like
        """
        if hasattr(param_source, "pvalue") and callable(param_source.pvalue):
            param_source = param_source.pvalue
        return sum(
            j.evaluate(
                param_source,
                x_namespace=x_namespace,
                exp_params=exp_params,
                **more_x_namespace,
            )
            for j in self
        )

    def value(self, *args):
        return self.as_pmath().value(*args)

    def copy(self):
        result = self.__class__(self)
        return result

    def __deepcopy__(self, memodict):
        result = self.__class__()
        import copy

        for i in self:
            result.append(copy.deepcopy(i, memodict))
        return result

    def _linear_plot_2d_data(
        self, p_getter, x_name, x_min, x_max, n_points=100, **other_namespace
    ):
        import numpy

        if hasattr(self, "plotting_namespace") and len(other_namespace) == 0:
            other_namespace = self.plotting_namespace
        x = numpy.linspace(x_min, x_max, n_points)
        y = self.evaluate(p_getter, {x_name: x}, **other_namespace)
        return x, y

    def linear_plot_2d(
        self,
        p_getter,
        x_name,
        x_min,
        x_max,
        n_points=100,
        *,
        xlabel=None,
        svg=True,
        header=None,
        **other_namespace,
    ):
        # Delayed evaluation mode...
        if p_getter is None:
            return lambda x: self.linear_plot_2d(
                x,
                x_name,
                x_min,
                x_max,
                n_points=n_points,
                xlabel=xlabel,
                svg=svg,
                header=header,
                **other_namespace,
            )

        # Active evaluation mode...
        from .._optional import pyplot as plt

        plt.clf()
        x, y = self._linear_plot_2d_data(
            p_getter, x_name, x_min, x_max, n_points, **other_namespace
        )
        if hasattr(self, "plotting_label"):
            plt.plot(x, y, label=self.plotting_label)
        else:
            plt.plot(x, y)
        if xlabel is None:
            plt.xlabel(x_name)
        else:
            plt.xlabel(xlabel)
        plt.tight_layout(pad=0.5)
        from ..util.plotting import plot_as_svg_xhtml

        if svg is True:
            svg = {}
        if svg or svg == {}:
            if header is not None:
                svg["header"] = header
            return plot_as_svg_xhtml(plt, **svg)
        else:
            plt.show()

    def _inplot_linear_plot_2d(
        self,
        plt,
        p_getter,
        x_name,
        x_min,
        x_max,
        n_points=100,
        *,
        xlabel=None,
        svg=True,
        header=None,
        **other_namespace,
    ):
        # Delayed evaluation mode...
        if p_getter is None:
            return lambda x: self._inplot_linear_plot_2d(
                plt,
                x,
                x_name,
                x_min,
                x_max,
                n_points=n_points,
                xlabel=xlabel,
                svg=svg,
                header=header,
                **other_namespace,
            )

        # Active evaluation mode...
        x, y = self._linear_plot_2d_data(
            p_getter, x_name, x_min, x_max, n_points, **other_namespace
        )
        if hasattr(self, "plotting_label"):
            plt.plot(x, y, label=self.plotting_label)
        else:
            plt.plot(x, y)

    def total_ordering_increasing(self):
        from toolz.itertoolz import sliding_window

        snaps = []
        for windows_size in range(2, len(self) - 1):
            for sub_p in sliding_window(windows_size, self):
                snaps.append(sub_p[0].param.lessthan(sub_p[-1].param))

        return snaps

    def as_pmath(self):
        for i in self:
            if i.data != "1":
                raise NotImplementedError(f"{type(self)} has non-unit data")
        from .linear_math import ParameterAdd

        if len(self) == 0:
            return 0
        elif len(self) == 1:
            return self[0].as_pmath()
        else:
            x = ParameterAdd(self[0].as_pmath(), self[1].as_pmath())
            for i in self[2:]:
                x = ParameterAdd(x, i.as_pmath())
            return x


class DictOfAlts(MutableMapping):
    _instance = None

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_private_" + name

    def __fresh(self, instance):
        newself = DictOfAlts(
            alts_validator=self._alts_validator,
            default_value=self._default_value,
            value_validator=self._value_validator,
        )
        newself._instance = instance
        setattr(instance, self.private_name, newself)
        return newself

    def __init__(
        self,
        mapping=None,
        alts_validator=None,
        default_value=1,
        value_validator=lambda i: True,
        **kwargs,
    ):
        self._map = {}
        self._value_validator = value_validator
        self._alts_validator = alts_validator
        self._default_value = default_value
        if mapping is None:
            mapping = {}
        for k, v in mapping.items():
            if not self._value_validator(v):
                raise ValueError(f"bad value [{k}]: {v}")
            self._map[k] = v
        for k, v in kwargs.items():
            self._map[k] = v

    def __get__(self, instance, instancetype):
        """
        Get attribute of instance.

        Parameters
        ----------
        instance : Any
                Instance of parent class that has `self` as a member.
        instancetype : class
                Class of `instance`.
        """
        if instance is None:
            return self
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        return newself

    def __set__(self, instance, value):
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        newself.__init__(
            value,
            alts_validator=self._alts_validator,
            default_value=self._default_value,
            value_validator=self._value_validator,
        )
        try:
            newself._instance.mangle()
        except AttributeError:
            pass

    def __delete__(self, instance):
        self.__set__(instance, None)

    def __getitem__(self, k):
        try:
            return self._map[k]
        except KeyError:
            if self._alts_validator is None or self._alts_validator(k):
                v = self._map[k] = self._default_value
                return v
            else:
                raise

    def __setitem__(self, k, v):
        if not self._value_validator(v):
            raise ValueError(f"bad value [{k}]: {v}")
        existing_value = self[k]
        if existing_value != v:
            self._map[k] = v
            try:
                self._instance.mangle()
            except AttributeError:
                pass

    def __delitem__(self, key):
        del self._map[key]
        _try_mangle(self._instance)

    def __iter__(self):
        return iter(self._map.keys())

    def __len__(self):
        return len(self._map)

    # def keys(self):
    # 	return self._map.keys()
    #
    # def items(self):
    # 	return self._map.items()
    #
    # def values(self):
    # 	return self._map.values()

    def copy(self):
        return type(self)(self)

    def __repr__(self):
        return f"{type(self).__name__}({repr(self._map)})"

    def to_dict(self):
        return dict(self)


class DictOfLinearFunction:
    _instance = None

    def __init__(self, mapping=None, alts_validator=None, **kwargs):
        self._map = {}
        if mapping is None:
            mapping = {}
        for k, v in mapping.items():
            self._map[k] = LinearFunction(v)
        for k, v in kwargs.items():
            try:
                self._map[k] = LinearFunction(v)
            except Exception:
                print(v)
                print(type(v))
                raise

        self._alts_validator = alts_validator
        # self._instance = None

    def __fresh(self, instance):
        newself = DictOfLinearFunction()
        newself._instance = instance
        setattr(instance, self.private_name, newself)
        return newself

    def __get__(self, instance, instancetype):
        """
        Get attribute of instance.

        Parameters
        ----------
        instance : Any
                Instance of parent class that has `self` as a member.
        instancetype : class
                Class of `instance`.
        """
        if instance is None:
            return self
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        return newself

    def __set__(self, instance, values):
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        newself.__init__(values)
        # _try_mangle_h(newself)
        try:
            newself._instance.mangle()
        except AttributeError:
            pass  # print(f"No Mangle L2: {err}")
        else:
            pass  # print(f"Yes Mangle L2: {newself._instance}")

    def __delete__(self, instance):
        self.__set__(instance, None)

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name

    def set_alts_validator(self, av):
        self._alts_validator = av

    def __getitem__(self, k):
        try:
            v = self._map[k]
            v.set_instance(self._instance)
            return v
        except KeyError:
            if self._alts_validator is None or self._alts_validator(k):
                v = self._map[k] = LinearFunction()
                v.set_instance(self._instance)
                return v
            else:
                raise

    def __setitem__(self, k, v):
        if isinstance(v, ParameterRef):
            v = v * DataRef("1")
        if isinstance(v, int) and v == 0:
            v = LinearFunction()
        elif isinstance(v, LinearComponent):
            v = LinearFunction([v])
        elif isinstance(v, list):
            v = LinearFunction(v)
        elif not isinstance(v, LinearFunction):
            raise TypeError(
                f"values in {self.__class__} can only have type LinearFunction, not {type(v)}"
            )
        if "with error" in repr(v):
            raise ValueError("found error here")
        v.set_instance(self._instance)
        self._map[k] = v
        _try_mangle(self._instance)

    def __delitem__(self, key):
        del self._map[key]
        _try_mangle(self._instance)

    def __iter__(self):
        return iter(self._map.keys())

    def __len__(self):
        return len(self._map)

    def keys(self):
        return self._map.keys()

    def items(self):
        return self._map.items()

    def values(self):
        return self._map.values()

    def copy(self):
        return type(self)(self)

    def __repr__(self):
        return f"{type(self).__name__}({repr(self._map)})"

    def __xml__(self):
        from xmle import Elem

        x = Elem("div")
        t = x.elem("table", style="margin-top:1px;")
        t.elem(
            "caption",
            text=f"<larch.{self.__class__.__name__}>",
            style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;",
        )
        if len(self):
            tr = t.elem("tr")
            tr.elem("th", text="alt")
            tr.elem("th", text="formula")
            for k, v in self._map.items():
                tr = t.elem("tr")
                tr.elem("td", text=str(k))
                try:
                    v_ = v.__xml__()
                except AttributeError:
                    tr.elem("td", text=str(v), style="text-align:left;")
                else:
                    tr.elem("td", style="text-align:left;") << v_
        else:
            tr = t.elem("tr")
            tr.elem("td", text="<empty>")
        return x

    def _repr_html_(self):
        return self.__xml__().tostring()


P = Ref_Gen(ParameterRef)
X = Ref_Gen(DataRef)



#!/usr/bin/env python
import hashlib
import os
import time
import zipfile


def _rec_split(s):
    rest, tail = os.path.split(s)
    if rest in ("", os.path.sep):
        return (tail,)
    return _rec_split(rest) + (tail,)


def _any_dot(s):
    for i in _rec_split(s):
        if len(i) > 0 and i[0] == ".":
            return True
    return False


def _zipdir(path, ziph, skip_dots=True, extra_layer=True, log=print):
    # ziph is zipfile handle
    keep_dots = not skip_dots
    for root, _dirs, files in os.walk(path):
        folder = os.path.basename(root)
        if keep_dots or not _any_dot(folder):
            log(f"zipping folder: {folder} in {root}")
            for file in files:
                if keep_dots or not _any_dot(file):
                    ziph.write(
                        os.path.join(root, file),
                        os.path.relpath(
                            os.path.join(root, file),
                            os.path.join(path, ".." if extra_layer else "."),
                        ),
                    )
        else:
            log(f"not zipping folder: {folder} in {root}")


def zipdir(
    source_dir, zip_file_name=None, skip_dots=True, extra_layer=False, log=print
):
    """
    Compress a folder into a zip file.

    Parameters
    ----------
    source_dir : Path-like
    zip_file_name : str
            If not given, uses the name of the sourcedir.
    skip_dots : bool, defaults True
            Ignore files and dirs that start with a dot.

    Returns
    -------
    str
            zip_file_name
    """
    if zip_file_name is None:
        if source_dir[-1] in ("/", "\\"):
            usepath = source_dir[:-1]
        else:
            usepath = source_dir
        zip_file_name = usepath + ".zip"
    with zipfile.ZipFile(zip_file_name, "w", zipfile.ZIP_DEFLATED) as zipf:
        _zipdir(source_dir, zipf, skip_dots=skip_dots, extra_layer=extra_layer, log=log)
    return zip_file_name


def zipmod(module, zip_file_name, skip_dots=True) -> None:
    """
    Create a zipfile from a module.

    Parameters
    ----------
    module
    zip_file_name
    skip_dots
    """
    with zipfile.ZipFile(zip_file_name, "w", zipfile.ZIP_DEFLATED) as zipf:
        _zipdir(module.__path__[0], zipf, skip_dots=skip_dots)


def zipmod_temp(module, skip_dots=True):
    import tempfile

    tempdir = tempfile.TemporaryDirectory()
    zip_file_name = os.path.join(tempdir.name, module.__name__ + ".zip")
    zipmod(module, zip_file_name, skip_dots=skip_dots)
    return zip_file_name, tempdir


def make_hash_file(fname):
    hash256 = hashlib.sha256()
    if fname[-3:] == ".gz":
        import gzip

        with gzip.open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash256.update(chunk)
        h = hash256.hexdigest()
        with open(fname[:-3] + ".sha256.txt", "w") as fh:
            fh.write(h)
    else:
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash256.update(chunk)
        h = hash256.hexdigest()
        with open(fname + ".sha256.txt", "w") as fh:
            fh.write(h)


def verify_hash_file(fname, hash_dir=None, max_retries=5):
    hash256 = hashlib.sha256()

    retries = 0
    while retries < max_retries:
        try:
            with open(fname, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    hash256.update(chunk)
        except PermissionError:
            import time

            time.sleep(5)
            retries += 1
        except Exception:
            raise
        else:
            break

    h = hash256.hexdigest()
    if hash_dir is None:
        with open(fname + ".sha256.txt") as fh:
            h_x = fh.read()
    else:
        with open(
            os.path.join(hash_dir, os.path.basename(fname) + ".sha256.txt")
        ) as fh:
            h_x = fh.read()
    if h != h_x:
        if hash_dir:
            raise ValueError(f"bad hash on {fname} with hash_dir={hash_dir}")
        else:
            raise ValueError(f"bad hash on {fname}")


def archive_subdir(destination_dir):
    import os
    import time

    destination_dir = os.path.join(destination_dir, time.strftime("%Y_%m_%d-%H_%M"))
    os.makedirs(destination_dir, exist_ok=True)
    return destination_dir


def gzip_dir(
    source_dir,
    patterns=("*.*",),
    make_hash=True,
    exclude=".sha256.txt",
    destination_dir=None,
    archive_subdir=False,
    remove_original=True,
    log=None,
):
    """Individually gzip every file matching patterns in source_dir."""
    import glob
    import gzip
    import os
    import shutil
    import time

    if destination_dir is None:
        destination_dir = source_dir
    else:
        os.makedirs(destination_dir, exist_ok=True)

    if archive_subdir:
        if isinstance(archive_subdir, str):
            destination_dir = os.path.join(destination_dir, archive_subdir)
        else:
            destination_dir = os.path.join(
                destination_dir, time.strftime("%Y_%m_%d-%H_%M")
            )
        os.makedirs(destination_dir, exist_ok=True)

    targets = set()

    for pattern in patterns:
        for f in glob.glob(os.path.join(source_dir, pattern)):
            if exclude in f:
                continue  # don't re-gzip the hash files by default
            targets.add(f)

    for f in targets:
        if make_hash:
            make_hash_file(f)
        f_relative = os.path.relpath(f, start=source_dir)

        if f[-3:] != ".gz":
            with open(f, "rb") as f_in:
                f_to = os.path.join(destination_dir, f_relative)
                with gzip.open(f_to + ".gz", "wb") as f_out:
                    if log is not None:
                        log(f"{f} --> {f_to}.gz")
                    shutil.copyfileobj(f_in, f_out)
            if remove_original:
                os.remove(f)


def gzip_archive(
    source_dir,
    patterns=("*.*",),
    make_hash=False,
    exclude=".sha256.txt",
    destination_dir=None,
    archive_subdir=True,
    remove_original=False,
    log=None,
):
    """
    Individually gzip every file matching patterns in source_dir.

    This function is the same as gzip_dir but with different default inputs.

    Parameters
    ----------
    source_dir
    patterns
    make_hash
    exclude
    destination_dir
    archive_subdir
    remove_original
    log
    """
    return gzip_dir(
        source_dir=source_dir,
        patterns=patterns,
        make_hash=make_hash,
        exclude=exclude,
        destination_dir=destination_dir,
        archive_subdir=archive_subdir,
        remove_original=remove_original,
        log=log,
    )


class Archiver:
    def __init__(self, *outdirs, log=None, handle=None):
        if handle is None:
            self._handle = time.strftime("%Y_%m_%d-%H_%M")
        else:
            self._handle = handle
        self.outdirs = []
        self.archive_dirs = []
        for od in outdirs:
            self.add_outdir(od)
        if log is None:
            from .logs.normal import log as _log

            self.log = _log.info
        elif log is False:
            self.log = None
        else:
            self.log = log

    def add_outdir(self, outdir):
        self.outdirs.append(outdir)
        d = os.path.join(outdir, self._handle)
        os.makedirs(d, exist_ok=True)
        self.archive_dirs.append(d)

    def gzip_by_glob(self, pattern, source_dir="."):
        for arch in self.archive_dirs:
            gzip_archive(
                source_dir=source_dir,
                patterns=(pattern,),
                destination_dir=arch,
                archive_subdir=False,
                log=self.log,
            )

    def zip_dir(self, source_dir, zip_file_name=None, **kwargs):
        if zip_file_name is None:
            zip_file_name = os.path.basename(source_dir) + ".zip"
        for arch in self.archive_dirs:
            zipdir(
                source_dir=source_dir,
                zip_file_name=os.path.join(arch, zip_file_name),
                log=self.log,
                **kwargs,
            )

    zip_by_dir = zip_dir
